use std::collections::HashSet;
use std::error::Error;
use std::io;
use imap::{Session};
use maildir::Maildir;
use native_tls::{TlsStream};
use std::net::TcpStream;
use std::path::{PathBuf};
use std::io::{Write};

pub async fn download_email_from_imap(maildir_path: PathBuf, imap_domain: String, username: String, password: String) -> Result<(), Box<dyn Error>>{
    let tls = native_tls::TlsConnector::builder().build().unwrap();
    let client = imap::connect((imap_domain.clone(), 993), imap_domain.clone(), &tls).unwrap();
    let mut imap_session = client
        .login(username, password)
        .map_err(|e| e.0)?;

    match list_imap_folders(&mut imap_session).await {
        Ok(folders) => {
            for folder in folders {
                println!("Start downloading {}", folder.clone());
                match fetch_and_store_emails(&mut imap_session, maildir_path.clone(), folder.clone()).await {
                    Ok(_) => println!("Emails fetched and stored successfully for {}.", folder),
                    Err(err) => eprintln!("Error during email fetching and storing: {}", err)
                }
            }
        },
        Err(e) => eprintln!("Error listing folders: {}", e),
    }
    
    // Logout from the IMAP session
    match imap_session.logout() {
        Ok(_) => println!("Logged out successfully."),
        Err(err) => eprintln!("Error logging out: {}", err),
    }

    Ok(())
}

pub async fn list_imap_folders(imap_session: &mut Session<TlsStream<TcpStream>>) -> imap::error::Result<Vec<String>> {
    // The function return all available folders(lists) on the IMAP server
    // Fetch the list of mailboxes (folders)
    let mailboxes = imap_session.list(None, Some("*"))?;
    let folder_names = mailboxes.iter().map(|mailbox| mailbox.name().to_string()).collect();

    Ok(folder_names)
}

pub async fn fetch_and_store_emails(imap_session: &mut Session<TlsStream<TcpStream>>, maildir_path: PathBuf, list: String) -> imap::error::Result<()> {
    // the function download all emails from the IMAP in the folder(list)
    imap_session.select(list.clone())?;
    let maildir = Maildir::from(maildir_path.clone().join(list.clone()));

    // creating files for maildir
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("new")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("cur")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("tmp")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    
    let uids_path = maildir_path.clone().join(".uids.txt");
    
    let uids_server = imap_session.uid_search("ALL").unwrap();
    let uids_local = read_local_uids(uids_path.clone()).unwrap();
    for uid in uids_server {
        if uids_local.contains(&uid.to_string()){
            continue
        }
        
        let messages = imap_session.uid_fetch(uid.to_string(), "(UID BODY[])")?;
        let message = messages.iter().next().ok_or(imap::error::Error::Bad("No messages found".to_string()))?;

        if let Some(body) = message.body() {
            let mail_file = maildir.store_new(body);
            match mail_file {
                Ok(_) => {},
                Err(e) => eprintln!("Failed to store email: {}", e),
            }
        } else {
            eprintln!("Failed to retrieve email body");
        }
    
        if let Some(uid) = message.uid {
            update_local_uids(uid.to_string(), uids_path.clone()).expect("Cannot write uid to .uids.txt");
        }

        println!("Downloaded message with UID {}", uid.to_string());
    }
    
    Ok(())
}


fn read_local_uids(uids_path: PathBuf) -> io::Result<HashSet<String>> {
    let contents = std::fs::read_to_string(uids_path).unwrap_or_default();
    Ok(contents.lines().map(|line| line.to_owned()).collect())
}

fn update_local_uids(uid: String, uids_path: PathBuf) -> io::Result<()> {
    // Open the file in append mode, or create it if it doesn't exist
    let mut file = std::fs::OpenOptions::new()
        .append(true)
        .create(true) // Ensure the file is created if it doesn't exist
        .open(uids_path)?;

    // Write the uid to the file
    writeln!(file, "{}", uid)?;

    Ok(())
}
